from _pydevd_bundle.pydevd_constants import get_frame, IS_CPYTHON, IS_64BIT_PROCESS, IS_WINDOWS, \
    IS_LINUX, IS_MAC, DebugInfoHolder, LOAD_NATIVE_LIB_FLAG, \
    ENV_FALSE_LOWER_VALUES, ForkSafeLock
from _pydev_bundle._pydev_saved_modules import thread, threading
from _pydev_bundle import pydev_log, pydev_monkey
import os.path
import platform
import ctypes
from io import StringIO
import sys
import traceback

_original_settrace = sys.settrace


class TracingFunctionHolder:
    '''This class exists just to keep some variables (so that we don't keep them in the global namespace).
    '''
    _original_tracing = None
    _warn = True
    _traceback_limit = 1
    _warnings_shown = {}


def get_exception_traceback_str():
    exc_info = sys.exc_info()
    s = StringIO()
    traceback.print_exception(exc_info[0], exc_info[1], exc_info[2], file=s)
    return s.getvalue()


def _get_stack_str(frame):

    msg = '\nIf this is needed, please check: ' + \
          '\nhttp://pydev.blogspot.com/2007/06/why-cant-pydev-debugger-work-with.html' + \
          '\nto see how to restore the debug tracing back correctly.\n'

    if TracingFunctionHolder._traceback_limit:
        s = StringIO()
        s.write('Call Location:\n')
        traceback.print_stack(f=frame, limit=TracingFunctionHolder._traceback_limit, file=s)
        msg = msg + s.getvalue()

    return msg


def _internal_set_trace(tracing_func):
    if TracingFunctionHolder._warn:
        frame = get_frame()
        if frame is not None and frame.f_back is not None:
            filename = frame.f_back.f_code.co_filename.lower()
            if not filename.endswith('threading.py') and not filename.endswith('pydevd_tracing.py'):

                message = \
                '\nPYDEV DEBUGGER WARNING:' + \
                '\nsys.settrace() should not be used when the debugger is being used.' + \
                '\nThis may cause the debugger to stop working correctly.' + \
                '%s' % _get_stack_str(frame.f_back)

                if message not in TracingFunctionHolder._warnings_shown:
                    # only warn about each message once...
                    TracingFunctionHolder._warnings_shown[message] = 1
                    sys.stderr.write('%s\n' % (message,))
                    sys.stderr.flush()

    if TracingFunctionHolder._original_tracing:
        TracingFunctionHolder._original_tracing(tracing_func)


_last_tracing_func_thread_local = threading.local()


def SetTrace(tracing_func):
    _last_tracing_func_thread_local.tracing_func = tracing_func

    if tracing_func is not None:
        if set_trace_to_threads(tracing_func, thread_idents=[thread.get_ident()], create_dummy_thread=False) == 0:
            # If we can use our own tracer instead of the one from sys.settrace, do it (the reason
            # is that this is faster than the Python version because we don't call
            # PyFrame_FastToLocalsWithError and PyFrame_LocalsToFast at each event!
            # (the difference can be huge when checking line events on frames as the
            # time increases based on the number of local variables in the scope)
            # See: InternalCallTrampoline (on the C side) for details.
            return

    # If it didn't work (or if it was None), use the Python version.
    set_trace = TracingFunctionHolder._original_tracing or sys.settrace
    set_trace(tracing_func)


def reapply_settrace():
    try:
        tracing_func = _last_tracing_func_thread_local.tracing_func
    except AttributeError:
        return
    else:
        SetTrace(tracing_func)


def replace_sys_set_trace_func():
    if TracingFunctionHolder._original_tracing is None:
        TracingFunctionHolder._original_tracing = sys.settrace
        sys.settrace = _internal_set_trace


def restore_sys_set_trace_func():
    if TracingFunctionHolder._original_tracing is not None:
        sys.settrace = TracingFunctionHolder._original_tracing
        TracingFunctionHolder._original_tracing = None


_lock = ForkSafeLock()


def _load_python_helper_lib():
    try:
        # If it's already loaded, just return it.
        return _load_python_helper_lib.__lib__
    except AttributeError:
        pass
    with _lock:
        try:
            return _load_python_helper_lib.__lib__
        except AttributeError:
            pass

        lib = _load_python_helper_lib_uncached()
        _load_python_helper_lib.__lib__ = lib
        return lib


def get_python_helper_lib_filename():
    # Note: we have an independent (and similar -- but not equal) version of this method in
    # `add_code_to_python_process.py` which should be kept synchronized with this one (we do a copy
    # because the `pydevd_attach_to_process` is mostly independent and shouldn't be imported in the
    # debugger -- the only situation where it's imported is if the user actually does an attach to
    # process, through `attach_pydevd.py`, but this should usually be called from the IDE directly
    # and not from the debugger).
    libdir = os.path.join(os.path.dirname(__file__), 'pydevd_attach_to_process')

    arch = ''
    if IS_WINDOWS:
        # prefer not using platform.machine() when possible (it's a bit heavyweight as it may
        # spawn a subprocess).
        arch = os.environ.get("PROCESSOR_ARCHITEW6432", os.environ.get('PROCESSOR_ARCHITECTURE', ''))

    if not arch:
        arch = platform.machine()
        if not arch:
            pydev_log.info('platform.machine() did not return valid value.')  # This shouldn't happen...
            return None

    if IS_WINDOWS:
        extension = '.dll'
        suffix_64 = 'amd64'
        suffix_32 = 'x86'

    elif IS_LINUX:
        extension = '.so'
        suffix_64 = 'amd64'
        suffix_32 = 'x86'

    elif IS_MAC:
        extension = '.dylib'
        suffix_64 = 'x86_64'
        suffix_32 = 'x86'

    else:
        pydev_log.info('Unable to set trace to all threads in platform: %s', sys.platform)
        return None

    if arch.lower() not in ('amd64', 'x86', 'x86_64', 'i386', 'x86'):
        # We don't support this processor by default. Still, let's support the case where the
        # user manually compiled it himself with some heuristics.
        #
        # Ideally the user would provide a library in the format: "attach_<arch>.<extension>"
        # based on the way it's currently compiled -- see:
        # - windows/compile_windows.bat
        # - linux_and_mac/compile_linux.sh
        # - linux_and_mac/compile_mac.sh

        try:
            found = [name for name in os.listdir(libdir) if name.startswith('attach_') and name.endswith(extension)]
        except:
            if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 1:
                # There is no need to show this unless debug tracing is enabled.
                pydev_log.exception('Error listing dir: %s', libdir)
            return None

        expected_name = 'attach_' + arch + extension
        expected_name_linux = 'attach_linux_' + arch + extension

        filename = None
        if expected_name in found:  # Heuristic: user compiled with "attach_<arch>.<extension>"
            filename = os.path.join(libdir, expected_name)

        elif IS_LINUX and expected_name_linux in found:  # Heuristic: user compiled with "attach_linux_<arch>.<extension>"
            filename = os.path.join(libdir, expected_name_linux)

        elif len(found) == 1:  # Heuristic: user removed all libraries and just left his own lib.
            filename = os.path.join(libdir, found[0])

        else:  # Heuristic: there's one additional library which doesn't seem to be our own. Find the odd one.
            filtered = [name for name in found if not name.endswith((suffix_64 + extension, suffix_32 + extension))]
            if len(filtered) == 1:  # If more than one is available we can't be sure...
                filename = os.path.join(libdir, found[0])

        if filename is None:
            pydev_log.info(
                'Unable to set trace to all threads in arch: %s (did not find a %s lib in %s).',
                arch, expected_name, libdir

            )
            return None

        pydev_log.info('Using %s lib in arch: %s.', filename, arch)

    else:
        # Happy path for which we have pre-compiled binaries.
        if IS_64BIT_PROCESS:
            suffix = suffix_64
        else:
            suffix = suffix_32

        if IS_WINDOWS or IS_MAC:  # just the extension changes
            prefix = 'attach_'
        elif IS_LINUX:  #
            prefix = 'attach_linux_'  # historically it has a different name
        else:
            pydev_log.info('Unable to set trace to all threads in platform: %s', sys.platform)
            return None

        filename = os.path.join(libdir, '%s%s%s' % (prefix, suffix, extension))

    if not os.path.exists(filename):
        pydev_log.critical('Expected: %s to exist.', filename)
        return None

    return filename


def _load_python_helper_lib_uncached():
    if (not IS_CPYTHON or sys.version_info[:2] > (3, 10)
            or hasattr(sys, 'gettotalrefcount') or LOAD_NATIVE_LIB_FLAG in ENV_FALSE_LOWER_VALUES):
        pydev_log.info('Helper lib to set tracing to all threads not loaded.')
        return None

    try:
        filename = get_python_helper_lib_filename()
        if filename is None:
            return None
        # Load as pydll so that we don't release the gil.
        lib = ctypes.pydll.LoadLibrary(filename)
        pydev_log.info('Successfully Loaded helper lib to set tracing to all threads.')
        return lib
    except:
        if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 1:
            # Only show message if tracing is on (we don't have pre-compiled
            # binaries for all architectures -- i.e.: ARM).
            pydev_log.exception('Error loading: %s', filename)
        return None


def set_trace_to_threads(tracing_func, thread_idents=None, create_dummy_thread=True):
    assert tracing_func is not None

    ret = 0

    # Note: use sys._current_frames() keys to get the thread ids because it'll return
    # thread ids created in C/C++ where there's user code running, unlike the APIs
    # from the threading module which see only threads created through it (unless
    # a call for threading.current_thread() was previously done in that thread,
    # in which case a dummy thread would've been created for it).
    if thread_idents is None:
        thread_idents = set(sys._current_frames().keys())

        for t in threading.enumerate():
            # PY-44778: ignore pydevd threads and also add any thread that wasn't found on
            # sys._current_frames() as some existing threads may not appear in
            # sys._current_frames() but may be available through the `threading` module.
            if getattr(t, 'pydev_do_not_trace', False):
                thread_idents.discard(t.ident)
            else:
                thread_idents.add(t.ident)

    curr_ident = thread.get_ident()
    curr_thread = threading._active.get(curr_ident)

    if curr_ident in thread_idents and len(thread_idents) != 1:
        # The current thread must be updated first (because we need to set
        # the reference to `curr_thread`).
        thread_idents = list(thread_idents)
        thread_idents.remove(curr_ident)
        thread_idents.insert(0, curr_ident)

    for thread_ident in thread_idents:
        # If that thread is not available in the threading module we also need to create a
        # dummy thread for it (otherwise it'll be invisible to the debugger).
        if create_dummy_thread:
            if thread_ident not in threading._active:

                class _DummyThread(threading._DummyThread):

                    def _set_ident(self):
                        # Note: Hack to set the thread ident that we want.
                        self._ident = thread_ident

                t = _DummyThread()
                # Reset to the base class (don't expose our own version of the class).
                t.__class__ = threading._DummyThread

                if thread_ident == curr_ident:
                    curr_thread = t

                with threading._active_limbo_lock:
                    # On Py2 it'll put in active getting the current indent, not using the
                    # ident that was set, so, we have to update it (should be harmless on Py3
                    # so, do it always).
                    threading._active[thread_ident] = t
                    threading._active[curr_ident] = curr_thread

                    if t.ident != thread_ident:
                        # Check if it actually worked.
                        pydev_log.critical('pydevd: creation of _DummyThread with fixed thread ident did not succeed.')

        # Some (ptvsd) tests failed because of this, so, leave it always disabled for now.
        # show_debug_info = 1 if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 1 else 0
        show_debug_info = 0

        # Hack to increase _Py_TracingPossible.
        # See comments on py_custom_pyeval_settrace.hpp
        proceed = thread.allocate_lock()
        proceed.acquire()

        def dummy_trace(frame, event, arg):
            return dummy_trace

        def increase_tracing_count():
            set_trace = TracingFunctionHolder._original_tracing or sys.settrace
            set_trace(dummy_trace)
            proceed.release()

        start_new_thread = pydev_monkey.get_original_start_new_thread(thread)
        start_new_thread(increase_tracing_count, ())
        proceed.acquire()  # Only proceed after the release() is done.
        proceed = None

        # Note: The set_trace_func is not really used anymore in the C side.
        set_trace_func = TracingFunctionHolder._original_tracing or sys.settrace

        lib = _load_python_helper_lib()
        if lib is None:  # This is the case if it's not CPython.
            pydev_log.info('Unable to load helper lib to set tracing to all threads (unsupported python vm).')
            ret = -1
        else:
            try:
                result = lib.AttachDebuggerTracing(
                    ctypes.c_int(show_debug_info),
                    ctypes.py_object(set_trace_func),
                    ctypes.py_object(tracing_func),
                    ctypes.c_uint(thread_ident),
                    ctypes.py_object(None),
                )
            except:
                if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 1:
                    # There is no need to show this unless debug tracing is enabled.
                    pydev_log.exception('Error attaching debugger tracing')
                ret = -1
            else:
                if result != 0:
                    pydev_log.info('Unable to set tracing for existing thread. Result: %s', result)
                    ret = result

    return ret




from typing import List, Tuple, Iterator, Optional, Any
import itertools
from robocorp_ls_core.protocols import IDocument
from robotframework_ls.impl.protocols import ICompletionContext, IRobotToken
import os

from robotframework_ls.impl.robot_constants import (
    COMMENT,
    HEADER_TOKENS,
    SETTING_TOKENS,
    NAME,
    KEYWORD_NAME,
    ARGUMENT,
    VARIABLE,
    KEYWORD,
    CONTROL_TOKENS,
    TESTCASE_NAME,
    ERROR,
    FATAL_ERROR,
    ROBOT_AND_TXT_FILE_EXTENSIONS,
    OPTION,
    ASSIGN,
)


TOKEN_TYPES = [
    "variable",
    "comment",
    # Custom added
    "header",  # entity.name.type.class.robot  -- *** Settings ***
    "setting",  # storage.type.setting.robot  -- Resource, Library
    "name",  # entity.other.inherited-class.robot   -- my.library.py
    "keywordNameDefinition",  # entity.name.function.robot
    "variableOperator",  # keyword.operator.variable.robot
    "keywordNameCall",  # meta.keyword.call.robot
    "settingOperator",  # keyword.operator.setting.robot
    "control",
    "testCaseName",
    "parameterName",
    "argumentValue",
    "error",
    "documentation",
]

TOKEN_MODIFIERS = [
    "declaration",
    "definition",
    "readonly",
    "static",
    "deprecated",
    "abstract",
    "async",
    "modification",
    "documentation",
    "defaultLibrary",
]

TOKEN_TYPE_TO_INDEX = {}
TOKEN_MODIFIER_TO_INDEX = {}

for i, value in enumerate(TOKEN_TYPES):
    TOKEN_TYPE_TO_INDEX[value] = i

for i, value in enumerate(TOKEN_MODIFIERS):
    TOKEN_MODIFIER_TO_INDEX[value] = 2 ** (i + 1)  # Modifiers use a bit mask.

del i
del value


def _extract_gherkin_token_from_keyword(
    keyword_token: IRobotToken,
) -> Tuple[Optional[IRobotToken], IRobotToken]:
    import re
    from robotframework_ls.impl.ast_utils import split_token_change_first

    result = re.match(
        r"^((Given|When|Then|And|But)\s+)", keyword_token.value, flags=re.IGNORECASE
    )
    if result:
        gherkin_token_length = len(result.group(1))
        return split_token_change_first(keyword_token, "control", gherkin_token_length)

    return None, keyword_token


def _extract_library_token_from_keyword(
    keyword_token: IRobotToken, scope: "_SemanticTokensScope"
) -> Tuple[Optional[IRobotToken], IRobotToken]:
    from robotframework_ls.impl.ast_utils import split_token_change_first

    if not "." in keyword_token.value:
        return None, keyword_token

    potential_candidates = _get_potential_library_names_from_keyword(
        keyword_token.value
    )

    for library_name in potential_candidates:
        if library_name in scope.imported_libraries:
            return split_token_change_first(keyword_token, "name", len(library_name))
    return None, keyword_token


def _get_potential_library_names_from_keyword(keyword_name: str) -> Iterator[str]:
    name_length = -1
    while True:
        name_length = keyword_name.find(".", name_length + 1)
        if name_length == -1:
            break
        library_name = keyword_name[:name_length].lower()
        yield library_name


def _iter_dependent_names(context: ICompletionContext) -> Iterator[str]:
    """
    Provides names which can be used as (library/resource) prefixes
    for keyword calls.

    Note: names returned are all lower-case as case should not be taken into
    account for matches.
    """
    dependency_graph = context.collect_dependency_graph()
    for library in dependency_graph.iter_all_libraries():
        name = library.name
        if name:
            library_name = os.path.basename(name)
            basename, ext = os.path.splitext(library_name)
            if ext == ".py":
                yield basename.lower()
            else:
                yield name.lower()

        alias = library.alias
        if alias:
            yield alias.lower()

    for resource_node, _ in dependency_graph.iter_all_resource_imports_with_docs():
        name = resource_node.name
        if name:
            resource_name = os.path.basename(name)
            basename, ext = os.path.splitext(resource_name)
            if ext in ROBOT_AND_TXT_FILE_EXTENSIONS:
                yield basename.lower()
            else:
                yield name.lower()


# See: https://code.visualstudio.com/api/language-extensions/semantic-highlight-guide#semantic-token-scope-map

RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX = {
    COMMENT: TOKEN_TYPE_TO_INDEX["comment"],
    NAME: TOKEN_TYPE_TO_INDEX["name"],
    KEYWORD_NAME: TOKEN_TYPE_TO_INDEX["keywordNameDefinition"],
    TESTCASE_NAME: TOKEN_TYPE_TO_INDEX["testCaseName"],
    KEYWORD: TOKEN_TYPE_TO_INDEX["keywordNameCall"],
    ARGUMENT: TOKEN_TYPE_TO_INDEX["argumentValue"],
    VARIABLE: TOKEN_TYPE_TO_INDEX["variable"],
    ERROR: TOKEN_TYPE_TO_INDEX["error"],
    FATAL_ERROR: TOKEN_TYPE_TO_INDEX["error"],
    OPTION: TOKEN_TYPE_TO_INDEX["argumentValue"],
}


for tok_type in HEADER_TOKENS:  # *** Settings ***, ...
    RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX[tok_type] = TOKEN_TYPE_TO_INDEX["header"]

for tok_type in SETTING_TOKENS:  # Library, Teardown, ...
    RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX[tok_type] = TOKEN_TYPE_TO_INDEX["setting"]

for tok_type in CONTROL_TOKENS:  # Library, Teardown, ...
    RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX[tok_type] = TOKEN_TYPE_TO_INDEX["control"]

# The assign will be tokenized and the '=' will be variableOperator.
RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX[ASSIGN] = TOKEN_TYPE_TO_INDEX["variableOperator"]

for key, val in list(RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX.items()):
    RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX[key.replace(" ", "_")] = val


VARIABLE_INDEX = TOKEN_TYPE_TO_INDEX["variable"]
ARGUMENT_INDEX = TOKEN_TYPE_TO_INDEX["argumentValue"]
VARIABLE_OPERATOR_INDEX = TOKEN_TYPE_TO_INDEX["variableOperator"]
SETTING_INDEX = TOKEN_TYPE_TO_INDEX["setting"]
PARAMETER_NAME_INDEX = TOKEN_TYPE_TO_INDEX["parameterName"]
DOCUMENTATION_INDEX = TOKEN_TYPE_TO_INDEX["documentation"]
CONTROL_INDEX = TOKEN_TYPE_TO_INDEX["control"]


def _tokenize_changing_argument_to_type(tokenize_variables_generator, use_type):
    from robotframework_ls.impl import ast_utils

    for tok in tokenize_variables_generator:
        if tok.type == ARGUMENT:
            yield ast_utils.copy_token_replacing(tok, type=use_type)
        else:
            yield tok


def _tokenize_variables(token: IRobotToken) -> Iterator[IRobotToken]:

    if token.type in (KEYWORD, ASSIGN):
        from robotframework_ls.impl import ast_utils

        # Hack because RF can't tokenize KEYWORD (it only tokenizes
        # some pre-defined types and KEYWORD is not there).

        if not token.value or "{" not in token.value:
            # Nothing to tokenize.
            return iter((token,))

        else:
            # Force ARGUMENT tokenization but show KEYWORD/ASSIGN for callers.
            t = ast_utils.copy_token_replacing(token, type=ARGUMENT)
            return _tokenize_changing_argument_to_type(
                t.tokenize_variables(), token.type
            )

    else:
        return token.tokenize_variables()


def semantic_tokens_range(context, range):
    return []


def _tokenize_token(
    node, use_token, scope: "_SemanticTokensScope"
) -> Iterator[Tuple[IRobotToken, int]]:
    if use_token.type in (use_token.EOL, use_token.SEPARATOR):
        # Fast way out for the most common tokens (which have no special handling).
        return

    from robotframework_ls.impl.variable_resolve import find_split_index

    from robotframework_ls.impl.ast_utils import (
        CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET,
        copy_token_replacing,
        split_token_change_first,
        is_node_with_expression_argument,
        iter_expression_tokens,
    )

    use_token_type = use_token.type
    in_documentation = False
    token_type_index: Optional[int]

    # Step 1: cast to KEYWORD if needed.
    if use_token_type == ARGUMENT:
        in_documentation = node.__class__.__name__ == "Documentation"

        if not in_documentation:

            in_expression = is_node_with_expression_argument(node)
            if scope.args_as_keywords_handler is not None:
                tok_type = scope.args_as_keywords_handler.next_tok_type(use_token)
                if tok_type == scope.args_as_keywords_handler.KEYWORD:
                    use_token_type = KEYWORD

                elif tok_type == scope.args_as_keywords_handler.EXPRESSION:
                    in_expression = True

                elif tok_type == scope.args_as_keywords_handler.CONTROL:
                    yield use_token, CONTROL_INDEX
                    return

            if in_expression:
                for token in iter_expression_tokens(use_token, "argumentValue"):
                    token_type_index = scope.get_index_from_internal_token_type(
                        token.type
                    )
                    if token_type_index is not None:
                        yield token, token_type_index
                return

    if use_token_type == NAME:
        if node.__class__.__name__ in CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET:
            use_token_type = KEYWORD

    if use_token.type != use_token_type:
        use_token = copy_token_replacing(use_token, type=use_token_type)

    if use_token_type == KEYWORD:
        token_keyword = use_token

        token_gherkin_prefix, token_keyword = _extract_gherkin_token_from_keyword(
            token_keyword
        )
        if token_gherkin_prefix:
            yield token_gherkin_prefix, scope.get_index_from_internal_token_type(
                token_gherkin_prefix.type
            )

        token_library_prefix, token_keyword = _extract_library_token_from_keyword(
            token_keyword, scope
        )
        if token_library_prefix:
            yield token_library_prefix, scope.get_index_from_internal_token_type(
                token_library_prefix.type
            )

        use_token = token_keyword

    try:
        iter_in = _tokenize_variables(use_token)
    except:
        if in_documentation:
            yield use_token, DOCUMENTATION_INDEX
        else:
            token_type_index = scope.get_index_from_rf_token_type(use_token_type)
            if token_type_index is not None:
                yield use_token, token_type_index
        return
    else:
        if use_token_type == ARGUMENT:
            first_token = next(iter_in)

            if in_documentation:
                equals_pos = -1
            else:
                equals_pos = find_split_index(first_token.value)
                if equals_pos != -1:
                    # Found an equals... let's check if it's not a 'catenate', which
                    # doesn't really accept parameters and just concatenates all...
                    value = node.get_value(use_token.KEYWORD)
                    if value and value.strip().lower() == "catenate":
                        equals_pos = -1

                        # Note: the best way to actually do this would be finding the
                        # reference to the keyword and then validating whether the
                        # keyword arguments match the expected name.
                        #
                        # For instance, a keyword call such as:
                        # Some Call     some arg = 22
                        #
                        # Should color `some arg =` differently only if the argument
                        # of `Some Call` is `some arg`, otherwise it should not color
                        # the argument as `same arg = 22` will be passed as a string
                        # to the positional argument 0 and not really a keyword parameter
                        # where `same arg` is set with value 22.
                        #
                        # Now, this requires a bit more tinkering with keyword caches
                        # and possibly semantic highlighting deltas to make sure the
                        # performance isn't negatively impacted by it.

            if equals_pos != -1:
                tok, first_token = split_token_change_first(
                    first_token, "parameterName", equals_pos
                )
                yield tok, PARAMETER_NAME_INDEX

                tok, first_token = split_token_change_first(
                    first_token, "variableOperator", 1
                )
                yield tok, VARIABLE_OPERATOR_INDEX

                # Add the remainder back.
                iter_in = itertools.chain(iter((first_token,)), iter_in)
            else:
                iter_in = itertools.chain(iter((first_token,)), iter_in)

        for token in iter_in:
            token_type_index = scope.get_index_from_rf_token_type(token.type)
            if token_type_index is not None:
                yield from _tokenized_args(
                    token, token_type_index, in_documentation, scope
                )


def _tokenize_subvars(var_token, token_type_index, scope):
    if "{" not in var_token.value:
        yield var_token, token_type_index
        return

    if var_token.value.startswith("{") and var_token.value.endswith("}"):
        from robotframework_ls.impl.ast_utils import split_token_in_3
        from robotframework_ls.impl.ast_utils import iter_expression_tokens

        # i.e.: We're dealing with an expression.
        first, second, third = split_token_in_3(
            var_token,
            "variableOperator",
            "argumentValue",
            "variableOperator",
            1,
            -1,
        )

        yield first, VARIABLE_OPERATOR_INDEX

        for token in iter_expression_tokens(second, "argumentValue"):
            token_type_index = scope.get_index_from_internal_token_type(token.type)
            if token_type_index is not None:
                yield token, token_type_index

        yield third, VARIABLE_OPERATOR_INDEX
        return

    from robotframework_ls.impl.ast_utils import RobotMatchTokensGenerator

    robot_match_generator = RobotMatchTokensGenerator(var_token, "variable")
    from robotframework_ls.impl.variable_resolve import iter_robot_variable_matches

    for robot_match, relative_index in iter_robot_variable_matches(var_token.value):
        for token in robot_match_generator.gen_tokens_from_robot_match(
            robot_match,
            relative_index,
        ):
            token_type_index = scope.get_index_from_internal_token_type(token.type)
            if token_type_index is not None:
                yield token, token_type_index

    for token in robot_match_generator.gen_default_type(len(var_token.value)):
        token_type_index = scope.get_index_from_internal_token_type(token.type)
        if token_type_index is not None:
            yield token, token_type_index


def _tokenized_args(token, token_type_index, in_documentation, scope):
    from robotframework_ls.impl.ast_utils import split_token_change_first
    from robotframework_ls.impl.ast_utils import split_token_change_second
    from robotframework_ls.impl.ast_utils import split_token_in_3

    if in_documentation and token_type_index == ARGUMENT_INDEX:
        # Handle the doc itself (note that we may also tokenize docs
        # to include variables).
        yield token, DOCUMENTATION_INDEX
        return

    if token_type_index == VARIABLE_INDEX:
        from robotframework_ls.impl.variable_resolve import robot_search_variable

        variable_match = robot_search_variable(token.value)
        if variable_match is None or variable_match.base is None:
            yield token, token_type_index
            return

        after_token = None
        if variable_match.end > 0 and variable_match.end < len(token.value):
            token, after_token = split_token_change_first(
                token, token.type, variable_match.end
            )

        if variable_match.start > 0:
            before_token, token = split_token_change_first(
                token, token.type, variable_match.start
            )
            yield before_token, token_type_index

        base = variable_match.base
        if base:
            i = token.value.find(base)
        else:
            i = token.value.find("{") + 1
        op_start_token, var_token, op_end_token = split_token_in_3(
            token,
            "variableOperator",
            token.type,
            "variableOperator",
            i,
            i + len(base),
        )

        yield op_start_token, VARIABLE_OPERATOR_INDEX
        yield from _tokenize_subvars(var_token, token_type_index, scope)

        if variable_match.items:
            for item in variable_match.items:
                token = op_end_token
                i = token.value.find(item)

                op_start_token, var_token, op_end_token = split_token_in_3(
                    token,
                    "variableOperator",
                    token.type,
                    "variableOperator",
                    i,
                    i + len(item),
                )

                yield op_start_token, VARIABLE_OPERATOR_INDEX
                yield from _tokenize_subvars(var_token, token_type_index, scope)

        yield op_end_token, VARIABLE_OPERATOR_INDEX

        if after_token:
            yield after_token, token_type_index
        return

    if (
        token_type_index == SETTING_INDEX
        and len(token.value) > 2
        and token.value[-1] == "]"
        and token.value[0] == "["
    ):
        # We want to do an additional tokenization on names to
        # convert '[Arguments]' to '[', 'Arguments', ']'
        op_start_token, token = split_token_change_first(token, "settingOperator", 1)
        yield op_start_token, VARIABLE_OPERATOR_INDEX

        var_token, op_end_token = split_token_change_second(
            token, "settingOperator", len(token.value) - 1
        )
        yield var_token, token_type_index
        yield op_end_token, VARIABLE_OPERATOR_INDEX

        return

    if token.type == OPTION:
        eq_i = token.value.index("=")
        if eq_i != -1:
            # Convert limit=10 to 'limit' '=' '10'
            var_start_token, token = split_token_change_first(
                token, "parameterName", eq_i
            )
            yield var_start_token, PARAMETER_NAME_INDEX

            var_token, var_end_token = split_token_change_second(
                token, "variableOperator", 1
            )
            yield var_token, VARIABLE_OPERATOR_INDEX
            yield var_end_token, ARGUMENT_INDEX
            return

    # Default case (just yield the current token/type).
    yield token, token_type_index


class _SemanticTokensScope:
    def __init__(self, context: ICompletionContext):
        # It's the same for all files.
        self.imported_libraries = set(_iter_dependent_names(context))

        # Note: it's set for the node and then reused for all the tokens in that same node.
        self.args_as_keywords_handler: Any = None

        self.get_index_from_rf_token_type = RF_TOKEN_TYPE_TO_TOKEN_TYPE_INDEX.get
        self.get_index_from_internal_token_type = TOKEN_TYPE_TO_INDEX.__getitem__


def semantic_tokens_full(context: ICompletionContext):
    try:
        ast = context.doc.get_ast()
    except:
        return []

    from robotframework_ls.impl import ast_utils

    monitor = context.monitor

    ret: List[int] = []
    append = ret.append

    last_line = 0
    last_column = 0

    scope = _SemanticTokensScope(context)
    for _stack, node in ast_utils.iter_all_nodes_recursive(ast):
        if monitor:
            monitor.check_cancelled()
        tokens = getattr(node, "tokens", None)
        if tokens:
            scope.args_as_keywords_handler = ast_utils.get_args_as_keywords_handler(
                node
            )

            for token in tokens:
                for token_part, token_type_index in _tokenize_token(node, token, scope):
                    lineno = token_part.lineno - 1
                    if lineno < 0:
                        lineno = 0
                    append(lineno - last_line)
                    if lineno != last_line:
                        last_column = token_part.col_offset
                        if last_column < 0:
                            last_column = 0
                        append(last_column)
                    else:
                        col_delta = token_part.col_offset - last_column
                        append(col_delta)
                        last_column += col_delta

                    append(token_part.end_col_offset - token_part.col_offset)  # len
                    append(token_type_index)
                    append(0)  # i.e.: no modifier
                    last_line = lineno

    return ret


def decode_semantic_tokens(
    semantic_tokens_as_int: List[int], doc: IDocument, stream=None
):
    ret: List[Tuple[str, str]] = []
    if not semantic_tokens_as_int:
        return ret

    ints_iter = iter(semantic_tokens_as_int)
    line = 0
    col = 0
    while True:
        try:
            line_delta = next(ints_iter)
        except StopIteration:
            return ret
        col_delta = next(ints_iter)
        token_len = next(ints_iter)
        token_type = next(ints_iter)
        token_modifier = next(ints_iter)
        line += line_delta
        if line_delta == 0:
            col += col_delta
        else:
            col = col_delta

        s = doc.get_line(line)[col : col + token_len]
        ret.append((s, TOKEN_TYPES[token_type]))
        if stream is not None:
            print(">>", s, "<<", file=stream)

            print(f"line: {line}", file=stream)
            print(f"col: {col}", file=stream)
            print(f"line_delta: {line_delta}", file=stream)
            print(f"col_delta: {col_delta}", file=stream)
            print(f"len: {token_len}", file=stream)
            print(f"type: {token_type}", file=stream)
            print(f"modifier: {token_modifier}", file=stream)
            print("", file=stream)



'''
Things this test checks:

- frame.f_trace is None when there are only regular breakpoints.

- The no-op tracing function is set by default (otherwise when set tracing functions have no effect).

- When stepping in, frame.f_trace must be set by the frame eval.

- When stepping over/return, the frame.f_trace must not be set on intermediate callers.

TODO:

- When frame.f_trace is set to the default tracing function, it'll become None again in frame
  eval mode if not stepping (if breakpoints weren't changed).

- The tracing function in the frames that deal with unhandled exceptions must be set when dealing
  with unhandled exceptions.

- The tracing function in the frames that deal with unhandled exceptions must NOT be set when
  NOT dealing with unhandled exceptions.

- If handled exceptions should be dealt with, the proper tracing should be set in frame.f_trace.
'''

import sys
from _pydevd_frame_eval import pydevd_frame_tracing


def check_with_no_trace():
    if False:
        print('break on check_with_trace')
    frame = sys._getframe()
    if frame.f_trace is not None:
        raise AssertionError('Expected %s to be None' % (frame.f_trace,))

    if sys.gettrace() is not pydevd_frame_tracing.dummy_tracing_holder.dummy_trace_func:
        raise AssertionError('Expected %s to be dummy_trace_func' % (sys.gettrace(),))


def check_step_in_then_step_return():
    frame = sys._getframe()
    f_trace = frame.f_trace
    if f_trace.__class__.__name__ != 'SafeCallWrapper':
        raise AssertionError('Expected %s to be SafeCallWrapper' % (f_trace.__class__.__name__,))

    check_with_no_trace()


def check_revert_to_dummy():
    check_with_no_trace()


if __name__ == '__main__':
    # Check how frame eval works.
    if sys.version_info[0:2] < (3, 6):
        raise AssertionError('Only available for Python 3.6 onwards. Found: %s' % (sys.version_info[0:1],))

    check_with_no_trace()  # break on global (step over)

    check_step_in_then_step_return()

    import pydevd_tracing
    import pydevd

    # This is what a remote attach would do (should revert to the frame eval mode).
    pydevd_tracing.SetTrace(pydevd.get_global_debugger().trace_dispatch)
    check_revert_to_dummy()

    print('TEST SUCEEDED!')



def method1():
    yield
    print('here')  # Break here


if __name__ == '__main__':
    # i.e.: make sure we create 2 frames with different frameIds.
    it1 = iter(method1())
    it2 = iter(method1())

    next(it1)  # resume first
    next(it2)  # resume second

    try:
        next(it1)  # finish first
    except StopIteration:
        pass
    try:
        next(it2)  # finish second
    except StopIteration:
        pass

    print('TEST SUCEEDED')



import os.path
import sys
from typing import Any


def _import_yaml():
    try:
        import yaml
    except ImportError:
        _parent_dir = os.path.dirname(__file__)
        _yaml_dir = os.path.join(_parent_dir, "libs", "yaml_lib")
        if not os.path.exists(_yaml_dir):
            raise RuntimeError("Expected: %s to exist." % (_yaml_dir,))
        sys.path.append(_yaml_dir)
        import yaml  # @UnusedImport


def load(stream) -> Any:
    _import_yaml()
    import yaml

    return yaml.safe_load(stream)


def dumps(contents: Any) -> str:
    _import_yaml()
    import yaml

    return yaml.safe_dump(contents)



"""
This module provides a way to create a service which will listen for changes
on folders on the filesystem and will provide notifications of those to listeners.
"""
import os
import sys
import argparse
import threading
import socket as socket_module
from typing import Optional, Tuple

__file__ = os.path.abspath(__file__)
if __file__.endswith((".pyc", ".pyo")):
    __file__ = __file__[:-1]

# ==============================================================================
# Starting up as main below
# ==============================================================================

LOG_FORMAT = "%(asctime)s UTC pid: %(process)d - %(threadName)s - %(levelname)s - %(name)s\n%(message)s\n\n"

_critical_error_log_file = os.path.join(
    os.path.expanduser("~"), "remote_fs_observer_critical.log"
)


def add_arguments(parser):
    parser.description = "Remote FS Observer"

    parser.add_argument(
        "--log-file",
        help="Redirect logs to the given file instead of writing to stderr (i.e.: c:/temp/my_log.log).",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity of log output (i.e.: -vv).",
    )


class ObserverProvider(object):
    def __init__(self):
        from robocorp_ls_core.watchdog_wrapper import IFSObserver

        self._observer: Optional[IFSObserver] = None

    @property
    def observer(self):
        return self._observer

    def initialize_observer(self, backend, extensions: Optional[Tuple[str, ...]]):
        from robocorp_ls_core import watchdog_wrapper

        assert self._observer is None
        self._observer = watchdog_wrapper.create_observer(
            backend, extensions=extensions
        )
        return self._observer


class _RemoteFSServer(threading.Thread):
    def __init__(
        self, socket: socket_module.socket, observer_provider: ObserverProvider
    ) -> None:
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamWriter
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamReader
        from typing import Dict
        from robocorp_ls_core.watchdog_wrapper import IFSWatch

        threading.Thread.__init__(self)
        self.name = "_RemoteFSServer"
        self.socket = socket

        self.writer: Optional[JsonRpcStreamWriter] = None
        self.reader: Optional[JsonRpcStreamReader] = None
        self._observer_provider: ObserverProvider = observer_provider
        self.on_change_id_to_watch: Dict[int, IFSWatch] = {}

    def _on_change(self, src_path, on_change_id):
        # Note: this will be called from the watcher thread.
        if not self.writer.write(
            {"command": "on_change", "on_change_id": on_change_id, "src_path": src_path}
        ):
            from robocorp_ls_core.watchdog_wrapper import IFSWatch

            # I.e.: socket communication appears to be dropped for this change_id,
            # so, stop tracking it.
            watch: IFSWatch = self.on_change_id_to_watch.pop(on_change_id, None)
            if watch is not None:
                watch.stop_tracking()

    def run(self):
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamWriter
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamReader

        s = self.socket
        write_to = s.makefile("wb")
        read_from = s.makefile("rb")

        w = JsonRpcStreamWriter(write_to, sort_keys=True)
        r = JsonRpcStreamReader(read_from)
        self.writer = w
        self.reader = r

        r.listen(self._on_read)

    def _on_read(self, msg):
        command = msg.get("command")
        if command == "initialize":
            from robocorp_ls_core.basic import exit_when_pid_exists

            parent_pid = msg["parent_pid"]
            if parent_pid not in (None, -1, 0):
                exit_when_pid_exists(parent_pid)

            backend = msg["backend"]
            extensions = msg["extensions"]
            if extensions is not None:
                extensions = tuple(extensions)

            self._observer_provider.initialize_observer(backend, extensions)
            # initialize requires an acknowledgement.
            self.writer.write({"command": "ack_initialize"})

        elif command == "initialize_connect":
            # just used to send an initialize acknowledgement.
            self.writer.write({"command": "ack_initialize"})

        elif command == "stop_tracking":
            from robocorp_ls_core.watchdog_wrapper import IFSWatch

            on_change_id = msg["on_change_id"]
            watch: IFSWatch = self.on_change_id_to_watch.pop(on_change_id, None)
            if watch is not None:
                watch.stop_tracking()

        elif command == "notify_on_any_change":
            from robocorp_ls_core.watchdog_wrapper import PathInfo

            on_change_id = msg["on_change_id"]
            paths = [PathInfo(p["path"], p["recursive"]) for p in msg["paths"]]
            extensions = msg["extensions"]
            if extensions is not None:
                extensions = tuple(extensions)

            observer = self._observer_provider.observer
            self.on_change_id_to_watch[on_change_id] = observer.notify_on_any_change(
                paths, self._on_change, call_args=(on_change_id,), extensions=extensions
            )
            # notify_on_any_change requires an acknowledgement.
            self.writer.write(
                {"command": "ack_notify_on_any_change", "on_change_id": on_change_id}
            )


def main():
    try:
        import robocorp_ls_core
    except ImportError:
        # Automatically add it to the path if this module is being executed.
        sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        import robocorp_ls_core  # @UnusedImport

    from robocorp_ls_core.robotframework_log import (
        configure_logger,
        log_args_and_python,
        get_logger,
    )

    parser = argparse.ArgumentParser()
    add_arguments(parser)

    original_args = sys.argv[1:]
    args = parser.parse_args(args=original_args)

    verbose = args.verbose
    log_file = args.log_file or ""

    if not log_file:
        # If not specified in args, also check the environment variables.
        log_file = os.environ.get("REMOTE_FS_OBSERVER_LOG_FILE", "")
        if log_file:
            verbose = 2

    configure_logger("remote_fs_observer", verbose, log_file)
    log = get_logger("robocorp_ls_core.remote_fs_observer__main__")
    log_args_and_python(log, original_args, robocorp_ls_core)

    # Ok, the initial structure is in place, let's start listening for connections.
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("127.0.0.1", 0))
    s.listen()

    # Print the port where the server is listening so that clients can connect to it.
    stdout_buffer = sys.__stdout__.buffer
    stdout_buffer.write(f"port: {s.getsockname()[1]}\n".encode("utf-8"))
    stdout_buffer.flush()

    observer_provider = ObserverProvider()
    while True:
        conn, _addr = s.accept()
        server = _RemoteFSServer(conn, observer_provider)
        server.start()


if __name__ == "__main__":
    try:
        if sys.version_info[0] <= 2:
            raise AssertionError(
                "Python 3+ is required for the RobotFramework Language Server.\nCurrent executable: "
                + sys.executable
            )
        main()
    except (SystemExit, KeyboardInterrupt):
        pass
    except:
        # Critical error (the logging may not be set up properly).
        import traceback

        # Print to file and stderr.
        with open(_critical_error_log_file, "a+") as stream:
            traceback.print_exc(file=stream)

        traceback.print_exc()



import pytest


def some_call():
    assert 0  # raise here


def test_example():
    some_call()  # stop here


if __name__ == '__main__':
    pytest.main([__file__, '--capture=no', '--noconftest'])
    print('TEST SUCEEDED!')



import itertools
from functools import partial
import threading
from robocorp_ls_core.basic import implements
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.protocols import (
    ILanguageServerClientBase,
    Sentinel,
    COMMUNICATION_DROPPED,
    IIdMessageMatcher,
    IMonitor,
    IRequestHandler,
)
from typing import Any, Union, Optional, List, Callable, Dict
from robocorp_ls_core.callbacks import Callback

log = get_logger(__name__)


class _MessageMatcher(object):
    def __init__(self):
        self.event = threading.Event()
        self.msg = None

    def notify(self, msg):
        # msg can be None if the communication was finished in the meanwhile.
        self.msg = msg
        self.event.set()


class _IdMessageMatcher(_MessageMatcher):
    def __init__(self, message_id):
        _MessageMatcher.__init__(self)

        self.message_id = message_id

    def __str__(self):
        return "IdMatcher(%s)" % (self.message_id,)

    __repr__ = __str__

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IIdMessageMatcher = check_implements(self)


class _PatternMessageMatcher(_MessageMatcher):
    def __init__(self, message_pattern):
        self.message_pattern = message_pattern
        _MessageMatcher.__init__(self)

    def matches(self, msg):
        for key, val in self.message_pattern.items():
            if msg.get(key) != val:
                return False
        return True

    def __str__(self):
        return "PatternMatcher(%s)" % (self.message_pattern,)

    __repr__ = __str__


def wait_for_message_matcher(
    message_matcher: IIdMessageMatcher,
    request_cancel: Callable[[Any], None],
    timeout: float,
    monitor: Optional[IMonitor],
) -> bool:
    """
    Note: in the case where the monitor is cancelled the cancel is automatically
    passed to the api too.

    :raises JsonRpcRequestCancelled:
        If the monitor is cancelled.

    :return True if the message completed in the available time and False otherwise.
    """
    import time

    if monitor is None:
        if message_matcher.event.wait(timeout):
            return True
    else:
        from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled

        try:
            maxtime = time.time() + timeout
            # We need to periodically check for the monitor, so, set up a busy loop that does that.

            # Check at least 20 times / second.
            delta = min(1 / 20.0, timeout)

            # Always do at least 1 check regardless of the time.
            if message_matcher.event.wait(delta):
                return True

            monitor.check_cancelled()

            while time.time() < maxtime:
                if message_matcher.event.wait(delta):
                    return True

                monitor.check_cancelled()

        except JsonRpcRequestCancelled:
            request_cancel(message_matcher.message_id)
            raise

    log.info(
        f"Request timed out: {message_matcher.message_id} after {timeout} (seconds)."
    )

    # If we got here, it timed out, so, we should cancel the request itself
    # (even if it the monitor wasn't cancelled).
    request_cancel(message_matcher.message_id)
    return False


def wait_for_message_matchers(
    message_matchers: List[Optional[IIdMessageMatcher]],
    monitor: Optional[IMonitor],
    request_cancel: Callable[[Any], None],
    timeout: float,
) -> List[IIdMessageMatcher]:
    """
    Note: in the case where the monitor is cancelled the cancel is automatically
    passed to the api too.

    :raises JsonRpcRequestCancelled:
        If the monitor is cancelled.
    """
    import time
    from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled

    accepted_message_matchers = []
    curtime = time.time()
    maxtime = curtime + timeout

    # Create a copy as we'll mutate it.
    message_matchers = message_matchers[:]

    while message_matchers:
        message_matcher = message_matchers.pop(0)

        if message_matcher is not None:
            # i.e.: wait X seconds and bail out if we can't get it.
            available_time = maxtime - time.time()
            if available_time <= 0:
                available_time = 0.0001  # Wait at least a bit for each.

            try:
                if wait_for_message_matcher(
                    message_matcher, request_cancel, available_time, monitor
                ):
                    accepted_message_matchers.append(message_matcher)
            except JsonRpcRequestCancelled:
                # If we got here, the monitor was cancelled... the current one was
                # already cancelled, so, go on and cancel the others too.
                for message_matcher in message_matchers:
                    if message_matcher is not None:
                        request_cancel(message_matcher.message_id)
                raise

    return accepted_message_matchers


class _ReaderThread(threading.Thread):
    def __init__(self, reader, on_received_message=None):
        threading.Thread.__init__(self)
        self.daemon = True
        self.reader = reader
        self._lock = threading.Lock()
        self._finished = False

        # Message matchers.
        self._id_message_matchers = {}  # msg id-> matcher
        self._pattern_message_matchers = {}  # id(matcher) -> matcher
        self._request_handlers: Dict[str, List[IRequestHandler]] = {}

        self.on_message = Callback()
        if on_received_message is not None:
            self.on_message.register(on_received_message)

    def run(self):
        try:
            self.reader.listen(self._on_message)
        finally:
            with self._lock:
                self._finished = True
                id_message_matchers = self._id_message_matchers
                self._id_message_matchers = {}

                pattern_message_matchers = self._pattern_message_matchers
                self._pattern_message_matchers = {}

            for message_matcher in id_message_matchers.values():
                message_matcher.notify(None)

            for message_matcher in pattern_message_matchers.values():
                message_matcher.notify(None)

    def _on_message(self, msg):
        try:
            self.on_message(msg)
        except:
            log.exception("Error processing: %s", msg)

        from robocorp_ls_core.options import Setup

        notify_matchers = []
        log.debug("Will handle read message: %s" % (msg,))
        with self._lock:
            for message_matcher in self._pattern_message_matchers.values():
                if message_matcher.matches(msg):
                    notify_matchers.append(message_matcher)

            for message_matcher in notify_matchers:
                del self._pattern_message_matchers[id(message_matcher)]

            if "id" in msg:
                message_matcher = self._id_message_matchers.pop(msg["id"], None)
                if message_matcher is not None:
                    notify_matchers.append(message_matcher)

            if Setup.options.DEBUG_MESSAGE_MATCHERS:
                log.debug(
                    "Notify matchers: %s\nRemaining id matchers: %s\nRemaining pattern matchers: %s"
                    % (
                        notify_matchers,
                        self._id_message_matchers,
                        self._pattern_message_matchers,
                    )
                )

        if self._request_handlers:
            method = msg.get("method")
            msg_id = msg.get("id")
            if method is not None and msg_id is not None:
                request_handlers = self._request_handlers.get(method)
                if request_handlers:
                    for request_handler in request_handlers:
                        try:
                            if request_handler(method, msg_id, msg.get("params")):
                                break
                        except:
                            log.exception(
                                "Error processing: %s in %s", msg, request_handler
                            )

        for message_matcher in notify_matchers:
            try:
                message_matcher.notify(msg)
            except:
                log.exception("Error processing: %s", msg)

    def obtain_pattern_message_matcher(self, message_pattern):
        """
        :param message_pattern:
            Obtains a matcher which will be notified when the given message pattern is
            returned.

        :return:
            None if it's already finished or the message matcher otherwise.
        """
        with self._lock:
            if self._finished:
                return None
            message_matcher = _PatternMessageMatcher(message_pattern)
            self._pattern_message_matchers[id(message_matcher)] = message_matcher
            return message_matcher

    def obtain_id_message_matcher(self, message_id) -> Optional[IIdMessageMatcher]:
        """
        :param message_id:
            Obtains a matcher which will be notified when the given message id is
            returned.

        :return:
            None if it's already finished or the message matcher otherwise.
        """
        with self._lock:
            if self._finished:
                return None
            message_matcher = _IdMessageMatcher(message_id)
            self._id_message_matchers[message_id] = message_matcher
            return message_matcher

    def register_request_handler(self, request_name: str, handler: IRequestHandler):
        self._request_handlers.setdefault(request_name, []).append(handler)


class LanguageServerClientBase(object):
    """
    A base implementation for talking with a process that implements the language
    server.
    """

    DEFAULT_TIMEOUT: Optional[
        int
    ] = None  # The default if not redefined is not having a timeout.

    def __init__(self, writer, reader, on_received_message=None):
        """

        :param JsonRpcStreamWriter writer:
        :param JsonRpcStreamReader reader:
        """
        self.writer = writer
        self.reader = reader

        t = _ReaderThread(reader, on_received_message=on_received_message)
        self._reader_thread = t
        self.on_message = t.on_message
        t.start()
        self.require_exit_messages = True
        self.next_id = partial(next, itertools.count())

    @implements(ILanguageServerClientBase.register_request_handler)
    def register_request_handler(self, message: str, handler: IRequestHandler) -> None:
        self._reader_thread.register_request_handler(message, handler)

    @implements(ILanguageServerClientBase.request_async)
    def request_async(self, contents: dict) -> Optional[IIdMessageMatcher]:
        message_id = contents["id"]
        message_matcher = self._reader_thread.obtain_id_message_matcher(message_id)
        if message_matcher is None:
            return None

        if not self.write(contents):
            return None

        return message_matcher

    @implements(ILanguageServerClientBase.request)
    def request(
        self,
        contents,
        timeout: Union[int, Sentinel, None] = Sentinel.USE_DEFAULT_TIMEOUT,
        default: Any = COMMUNICATION_DROPPED,
    ):
        """
        :param contents:
        :param timeout:
        :return:
            The returned message if everything goes ok.
            `default` if the communication dropped in the meanwhile and timeout was None.

        :raises:
            TimeoutError if the timeout was given and no answer was given at the available time
            (including if the communication was dropped).
        """
        if timeout is Sentinel.USE_DEFAULT_TIMEOUT:
            timeout = self.DEFAULT_TIMEOUT

        message_id = contents["id"]
        message_matcher = self._reader_thread.obtain_id_message_matcher(message_id)
        if message_matcher is None:
            if timeout:
                raise TimeoutError(
                    "Request timed-out (%s) - no message matcher: %s"
                    % (timeout, contents)
                )
            return default

        if not self.write(contents):
            if timeout:
                raise TimeoutError(
                    "Request timed-out (%s) - no write: %s" % (timeout, contents)
                )
            return default

        if not message_matcher.event.wait(timeout=timeout):
            raise TimeoutError("Request timed-out (%s): %s" % (timeout, contents))

        return message_matcher.msg

    @implements(ILanguageServerClientBase.obtain_pattern_message_matcher)
    def obtain_pattern_message_matcher(self, message_pattern):
        return self._reader_thread.obtain_pattern_message_matcher(message_pattern)

    @implements(ILanguageServerClientBase.obtain_id_message_matcher)
    def obtain_id_message_matcher(self, message_id):
        return self._reader_thread.obtain_id_message_matcher(message_id)

    @implements(ILanguageServerClientBase.write)
    def write(self, contents):
        return self.writer.write(contents)

    @implements(ILanguageServerClientBase.shutdown)
    def shutdown(self):
        self.write({"jsonrpc": "2.0", "id": self.next_id(), "method": "shutdown"})

    @implements(ILanguageServerClientBase.exit)
    def exit(self):
        self.write({"jsonrpc": "2.0", "id": self.next_id(), "method": "exit"})



r'''
Creating the needed environments for creating the pre-compiled distribution on Windows:

See:

build_tools\pydevd_release_process.txt

for building binaries/release process.
'''

from __future__ import unicode_literals
import os
import subprocess
import sys

miniconda_envs = os.getenv('MINICONDA_ENVS', r'C:\bin\Miniconda3\envs')

python_installations = [

    r'%s\py36_32\python.exe' % miniconda_envs,
    r'%s\py37_32\python.exe' % miniconda_envs,
    r'%s\py38_32\python.exe' % miniconda_envs,
    r'%s\py39_32\python.exe' % miniconda_envs,
    r'%s\py310_32\python.exe' % miniconda_envs,

    r'%s\py36_64\python.exe' % miniconda_envs,
    r'%s\py37_64\python.exe' % miniconda_envs,
    r'%s\py38_64\python.exe' % miniconda_envs,
    r'%s\py39_64\python.exe' % miniconda_envs,
    r'%s\py310_64\python.exe' % miniconda_envs,
]

root_dir = os.path.dirname(os.path.dirname(__file__))


def list_binaries():
    for f in os.listdir(os.path.join(root_dir, '_pydevd_bundle')):
        if f.endswith('.pyd'):
            yield f


def extract_version(python_install):
    return python_install.split('\\')[-2][2:]


def main():
    from generate_code import generate_dont_trace_files
    from generate_code import generate_cython_module

    # First, make sure that our code is up to date.
    generate_dont_trace_files()
    generate_cython_module()

    for python_install in python_installations:
        assert os.path.exists(python_install), '%s does not exist.' % (python_install,)

    from build import remove_binaries
    remove_binaries(['.pyd'])

    for f in list_binaries():
        raise AssertionError('Binary not removed: %s' % (f,))

    for i, python_install in enumerate(python_installations):
        print()
        print('*' * 80)
        print('*' * 80)
        print()
        new_name = 'pydevd_cython_%s_%s' % (sys.platform, extract_version(python_install))
        args = [
            python_install, os.path.join(root_dir, 'build_tools', 'build.py'), '--no-remove-binaries', '--target-pyd-name=%s' % new_name, '--force-cython']
        if i != 0:
            args.append('--no-regenerate-files')
        name_frame_eval = 'pydevd_frame_evaluator_%s_%s' % (sys.platform, extract_version(python_install))
        args.append('--target-pyd-frame-eval=%s' % name_frame_eval)
        print('Calling: %s' % (' '.join(args)))

        env = os.environ.copy()
        python_exe_dir = os.path.dirname(python_install)
        env['PATH'] = env['PATH'] + ';' + os.path.join(python_exe_dir, 'DLLs') + ';' + os.path.join(python_exe_dir, 'Library', 'bin')
        subprocess.check_call(args, env=env)


if __name__ == '__main__':
    main()

# To run do:
# See:
#
# build_tools\pydevd_release_process.txt
#
# for building binaries/release process.



import sys


def method3():
    raise IndexError('foo')  # raise indexerror line


def method2():
    return method3()  # reraise on method2


def method1():
    try:
        method2()  # handle on method1
    except:
        pass  # Ok, handled
    assert '__exception__' not in sys._getframe().f_locals


if __name__ == '__main__':
    method1()
    print('TEST SUCEEDED!')



'''
@author Fabio Zadrozny
'''
import sys
import unittest
import socket
import urllib
import pytest
import pycompletionserver


IS_JYTHON = sys.platform.find('java') != -1
DEBUG = 0

def dbg(s):
    if DEBUG:
        sys.stdout.write('TEST %s\n' % s)

@pytest.mark.skipif(not IS_JYTHON, reason='Jython related test')
class TestJython(unittest.TestCase):

    def test_it(self):
        dbg('ok')

    
    def test_message(self):
        t = pycompletionserver.CompletionServer(0)
        t.exit_process_on_kill = False

        l = []
        l.append(('Def', 'description'  , 'args'))
        l.append(('Def1', 'description1', 'args1'))
        l.append(('Def2', 'description2', 'args2'))

        msg = t.processor.format_completion_message('test_jyserver.py', l)

        self.assertEqual('@@COMPLETIONS(test_jyserver.py,(Def,description,args),(Def1,description1,args1),(Def2,description2,args2))END@@', msg)

        l = []
        l.append(('Def', 'desc,,r,,i()ption', ''))
        l.append(('Def(1', 'descriptio(n1', ''))
        l.append(('De,f)2', 'de,s,c,ription2', ''))
        msg = t.processor.format_completion_message(None, l)
        expected = '@@COMPLETIONS(None,(Def,desc%2C%2Cr%2C%2Ci%28%29ption, ),(Def%281,descriptio%28n1, ),(De%2Cf%292,de%2Cs%2Cc%2Cription2, ))END@@'

        self.assertEqual(expected, msg)

    
    def test_completion_sockets_and_messages(self):
        dbg('test_completion_sockets_and_messages')
        t, socket = self.create_connections()
        self.socket = socket
        dbg('connections created')

        try:
            #now that we have the connections all set up, check the code completion messages.
            msg = urllib.quote_plus('math')

            toWrite = '@@IMPORTS:%sEND@@' % msg
            dbg('writing' + str(toWrite))
            socket.send(toWrite)  #math completions
            completions = self.read_msg()
            dbg(urllib.unquote_plus(completions))

            start = '@@COMPLETIONS('
            self.assertTrue(completions.startswith(start), '%s DOESNT START WITH %s' % (completions, start))
            self.assertTrue(completions.find('@@COMPLETIONS') != -1)
            self.assertTrue(completions.find('END@@') != -1)


            msg = urllib.quote_plus('__builtin__.str')
            toWrite = '@@IMPORTS:%sEND@@' % msg
            dbg('writing' + str(toWrite))
            socket.send(toWrite)  #math completions
            completions = self.read_msg()
            dbg(urllib.unquote_plus(completions))

            start = '@@COMPLETIONS('
            self.assertTrue(completions.startswith(start), '%s DOESNT START WITH %s' % (completions, start))
            self.assertTrue(completions.find('@@COMPLETIONS') != -1)
            self.assertTrue(completions.find('END@@') != -1)



        finally:
            try:
                self.send_kill_msg(socket)


                while not t.ended:
                    pass  #wait until it receives the message and quits.


                socket.close()
            except:
                pass

    def get_free_port(self):
        from _pydev_bundle.pydev_localhost import get_socket_name
        return get_socket_name(close=True)[1]

    def create_connections(self):
        '''
        Creates the connections needed for testing.
        '''
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((pycompletionserver.HOST, 0))
        server.listen(1)  #socket to receive messages.

        from thread import start_new_thread
        t = pycompletionserver.CompletionServer(server.getsockname()[1])
        t.exit_process_on_kill = False

        start_new_thread(t.run, ())

        sock, _addr = server.accept()

        return t, sock

    def read_msg(self):
        msg = '@@PROCESSING_END@@'
        while msg.startswith('@@PROCESSING'):
            msg = self.socket.recv(1024)
            if msg.startswith('@@PROCESSING:'):
                dbg('Status msg:' + str(msg))

        while msg.find('END@@') == -1:
            msg += self.socket.recv(1024)

        return msg

    def send_kill_msg(self, socket):
        socket.send(pycompletionserver.MSG_KILL_SERVER)




# Run for jython in command line:
# c:\bin\jython2.7.0\bin\jython.exe -m py.test tests\test_jyserver.py
